function vsf = vision_system_functions


    vsf.normalize_rgb_image = @normalize_rgb_image;
    
    vsf.find_red_pixels_from_rgb = @find_red_pixels_from_rgb;
    vsf.find_red_pixels_from_hsv = @find_red_pixels_from_hsv;
    
    vsf.countour_filter = @countour_filter;

    vsf.extract_feature_points_from_countours = @extract_feature_points_from_countours;
    
    vsf.find_target = @find_target;
    
    vsf.estimate_cam_position = @estimate_cam_position;
    
    vsf.zyx_rot = @zyx_rot;
    
    
    vsf.plot_feature_points = @plot_feature_points;
    vsf.show_contours = @show_contours;
end


%% Normalize RGB image

function I_rgb_norm = normalize_rgb_image(I_rgb)
    I_rgb = double(I_rgb);
    [V, U, ~] = size(I_rgb);
    
    I_rgb_norm = zeros(V, U, 3);
    for v = 1 : V
        for u = 1 : U
            R = I_rgb(v, u, 1);
            G = I_rgb(v, u, 2);
            B = I_rgb(v, u, 3);
            rgb_sum = (R + G + B);
            I_rgb_norm(v, u, 1) = R / rgb_sum;
            I_rgb_norm(v, u, 2) = G / rgb_sum;
            I_rgb_norm(v, u, 3) = B / rgb_sum;
        end
    end
end

%% Find red Pixels
function I_bin = find_red_pixels_from_rgb(I_rgb, T_1, T_2)
    [V, U, ~] = size(I_rgb);
    
    I_bin = zeros(V, U);
    
    for v = 1 : V
        for u = 1 : U
            rgb_vec = I_rgb(v, u, :);
            condition_1 = (rgb_vec(1) > (T_1 * rgb_vec(2))) && (rgb_vec(1) > (T_1 * rgb_vec(3)));
            condition_2 = rgb_vec(1) >= T_2 * (rgb_vec(2) + rgb_vec(3));
            if (condition_1 && condition_2) 
                I_bin(v, u) = 1;
            else
                I_bin(v, u) = 0;
            end
            
        end 
    end 
    I_bin = logical(I_bin);
    
end

function I_bin = find_red_pixels_from_hsv(I_hsv, T_H, T_S, T_V)
    h_layer = I_hsv(:, :, 1);
    s_layer = I_hsv(:, :, 2);
    v_layer = I_hsv(:, :, 3);
    
    h_red_mask = (h_layer <= T_H(1)) | (h_layer >= T_H(2));
    s_red_mask = (s_layer >= T_S(1)) & (s_layer <= T_S(2));
    v_red_mask = (v_layer >= T_V(1)) & (v_layer <= T_V(2));
    
    I_bin = h_red_mask & s_red_mask & v_red_mask;
    I_bin = logical(I_bin);
end

%% Contour filter
function filtered_contours = countour_filter(raw_contours, U, V)
    T_c_blob = 100;
    n_contours = length(raw_contours);
    
    % Loop over all countours and count up how many
    % we are going to keep.
    keep = 0;
    for n = 1 : n_contours
        contour = raw_contours(n);
        u = contour{1}(:,2);
        v = contour{1}(:,1);
        if (length(raw_contours{n}) > T_c_blob) && (is_circle(u, v, U, V))
            keep = keep + 1;
        end
    end
    
    filtered_contours = cell(keep, 1);
    n_contours = length(filtered_contours);
    keep = 1;
    for n = 1 : n_contours
        contour = raw_contours(n);
        u = contour{1}(:,2);
        v = contour{1}(:,1);
        if (length(raw_contours{n}) > T_c_blob) && (is_circle(u, v, U, V))
            filtered_contours{keep} = raw_contours{n};
            keep = keep + 1;
        end
    end    
    
    function accept = is_circle(u, v, U, V)

        k_1 = -0.5416; k_2 = 0.0656;
        k_3 = 0.0079; k_4 = 5 * k_3;

        alpha_min = 0.1 / 100;
        alpha_max = 20 / 100;

        c = length(u);

        d_u = abs(max(u) - min(u));
        d_v = abs(max(v) - min(v));

        d = (d_u + d_v) / 2;

        A = pi * (d / 2)^2;
        A_image = U * V;

        condition_1 = ((k_1 + pi) <= (c / d)) &&...
                      ((k_2 + pi) >= (c / d));

        condition_2 = ((k_3 + (1 / (4 * pi))) <= (A / c^2)) &&... 
                      ((k_4 + (1 / (4 * pi))) >= (A / c^2));

        condition_3 = ((alpha_min * A_image) <= A) &&... 
                      ((alpha_max * A_image) >= A);

        if ((condition_1 && condition_2) && condition_3)
            accept = true;
        else
            accept = false;
        end

    end
    
end

%% Extract feature points
function [u_min, u_max, v_min, v_max] = extract_feature_points_from_countours(filtered_contours)
    N = length(filtered_contours);
    
    u_min = zeros(1, N); u_max = zeros(1, N); 
    v_min = zeros(1, N); v_max = zeros(1, N); 
    
    for i = 1 : N
        contour = filtered_contours(i);
        u_temp = contour{1}(:, 2);
        v_temp = contour{1}(:, 1);
        u_min(i) = min(u_temp); u_max(i) = max(u_temp);
        v_min(i) = min(v_temp); v_max(i) = max(v_temp);

    end
end

function [u_target, v_target] = find_target(u_min, u_max, v_min, v_max)
    [~, N] = size(u_min);
    u_target = 0; v_target = 0;
    for i = 1 : N
        u_target = u_target + ((u_min(i) + u_max(i)));
        v_target = v_target + ((v_min(i) + v_max(i)));
    end
    u_target = round((u_target) / (2 * N));
    v_target = round(v_target / (2 * N));
end

%% Position Estimation

function P_c = estimate_cam_position(f_pix, u_min, u_max, v_min, v_max, u_0, v_0, u_target, v_target, d_real)

    Z_c_u = estimate_vertical_cam_position(f_pix, u_min, u_max, d_real);
    Z_c_v = estimate_vertical_cam_position(f_pix, v_min, v_max, d_real);
    
    Z_c = (Z_c_u + Z_c_v) / 2;
    X_c = Z_c * ((u_target - u_0) / (f_pix));
    Y_c = Z_c * ((v_target - v_0) / (f_pix));
    
    P_c = [X_c Y_c Z_c]';
    
    % Internal function
    function Z = estimate_vertical_cam_position(f_pix, dist_min, dist_max, d_real)
        ratio = 0;
        
        [~, N] = size(dist_min);

        for i = 1 : N - 1
            ratio = ratio + (1 / abs(dist_min(i) - dist_min(i + 1))); 
            ratio = ratio + (1 / abs(dist_max(i) - dist_max(i + 1))); 
        end

        Z = ((f_pix * d_real) / ((2*(N - 1)))) * ratio;
    end
end

%% Rotations
% Roll
function R_x_phi = principal_rot_x(phi)
    R_x_phi = [1     0            0;
               0 cos(phi) -sin(phi);
               0 sin(phi) cos(phi)];
end

% Pitch
function R_y_theta = principal_rot_y(theta)
    R_y_theta = [cos(theta) 0 sin(theta);
                0           1         0;
                -sin(theta) 0 cos(theta)]; 
end

% Yaw
function R_z_psi = principal_rot_z(psi)
   R_z_psi = [cos(psi) -sin(psi) 0;
              sin(psi) cos(psi)  0;
              0           0      1]; 
end

% Rotate according to the zyx convention
function R_zyx = zyx_rot(Theta)
    R_zyx = principal_rot_z(Theta(3)) * ...
            principal_rot_y(Theta(2)) * ...
            principal_rot_x(Theta(1));
end

%% Helper functions (not part of the actual algorithm)

% Adapted from:
% https://se.mathworks.com/help/images/ref/bwboundaries.html
function show_contours(contours, I_bin)
    imshow(I_bin); hold on;
    colors = ['y' 'g' 'r' 'c' 'm' 'y'];
    for i = 1 : length(contours)
        boundary = contours{i};
        color_index = mod(i, length(colors)) + 1; 
        plot(boundary(:,2), boundary(:, 1),...
           colors(color_index),'LineWidth', 2);
    end
end

function plot_feature_points(u_min, u_max, v_min, v_max, u_target, v_target, u_0, v_0,  I_bin)
    [~, N] = size(u_min);
    imshow(I_bin); hold on;
    marker_size = 10;
    leg = zeros(1, 6);
    
    for i = 1 : N
        leg(1) = plot(u_min(i), v_target , 'm+', 'MarkerSize', marker_size, 'DisplayName', 'u_{min}');
        leg(2) = plot(u_max(i), v_target , 'r+', 'MarkerSize', marker_size, 'DisplayName', 'u_{max}');
        
        
        leg(3) = plot(u_target, v_min(i), 'b+', 'MarkerSize', marker_size, 'DisplayName', 'v_{min}');
        leg(4) = plot(u_target, v_max(i), 'g+', 'MarkerSize', marker_size, 'DisplayName', 'v_{max}');
        
    end
    leg(5) = plot(u_target, v_target, 'ro', 'MarkerSize', marker_size, 'DisplayName', 'Target Center');
    leg(6) = plot(u_0, v_0, 'c*', 'MarkerSize', marker_size, 'DisplayName', 'Optical Center');
    
    legend(leg);
end






